/*
 * Perfmon Performance Monitor
 *
 * Copyright 1985, Massachusetts Institute of Technology
 * Copyright 1989, PCS Computer Systeme GmbH, West Germany
 * Copyright 1994, Sterling Software @ NASA-Ames Research Center
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of PCS and Sterling Software not be used in advertising or
 * publicity pertaining to distribution of the software without specific,
 * written prior permission.  PCS and Sterling Software makes no representations about the
 * suitability of this software for any purpose.  It is provided "as is"
 * without express or implied warranty.
 *
 * PCS & STERLING SOFTWARE DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL PCS & STERLING SOFTWARE
 * BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION
 * OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN 
 * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 * Original Author:  Emanuel Jay Berkenbilt, MIT Project Athena
 * Author:           Thomas A. Baghli, PCS Computer Systeme GmbH, West Germany
 *                   tom@meepmeep.pcs.com
 * 1994 Revision
 * Author:           Roger Smith, Sterling Software @ NASA-Ames Research Center
 *                   Moffett Field, California, rsmith@proteus.arc.nasa.gov
 */

/* This file contains only system functions - that is the functions that
 * get the information the performance monitor is monitoring.  No calls
 * to any X routines should be made here.  The reason for doing this is
 * so that as the X toolkit becomes available and the X window system 
 * improves no changes will have to be made to this file, and as this
 * program is made available for a new type of machine, only this file
 * will need to be changed.
 */
#include <X11/IntrinsicP.h>

#include "system.h"

#include <stdio.h>
#include <strings.h>

#ifdef PCS
#include <sys/types.h>
#include <sys/sysinfo.h>
#endif /* PCS */
#include <sys/param.h>
#include <sys/socket.h>
#ifndef PCS
#include <sys/vm.h>
#include <sys/dk.h>
#endif /* !PCS */
#include <nlist.h>
/*#include <sys/buf.h>*/

#ifdef vax
#include <vaxuba/ubavar.h>
#include <vaxmba/mbavar.h>
#endif vax
#ifdef sun
#include <sundev/mbvar.h>
#endif sun
#ifdef ibm032	/* IBM RT/PC */
#include <caio/ioccvar.h>
#endif ibm032

#include <net/if.h>
#include <netinet/in.h>
#include <sys/file.h>

#ifndef TRUE
#define TRUE 1
#define FALSE 0
#endif

#define USER		0
#define NICE		1
#define SYS		2
#define IDLE		3

#include "is.h"

extern int errno, sys_nerr;
extern char *sys_errlist[];
#define strerror(e) (((e) >= 0 && (e) < sys_nerr) ? sys_errlist[(e)] : "Unknown error")

#define VMUNIX  "/vmunix"
#define KMEM    "/dev/kmem"
#define MEM     "/dev/mem"

/* declarations for load_avg */
#include "loadavg.h"
#define NUM_AVERAGES 3

/* declarations for cpu states */
#include <machine/cpu.h>
#include <machine/pcb.h>	 /* gets cpudata.h for definition of cpudata struct */

/* define what weighted cpu is.  */
#define weighted_cpu(pct, pp) ((pp)->p_time == 0 ? 0.0 : \
                         ((pct) / (1.0 - exp((pp)->p_time * logcpu))))

extern struct _rpcNnfs rpcNnfs;
extern Widget perfmon[NUM_GRAPHS];

struct packet {
    int input, output, collisions;
};
static struct packet packets, old_packets;

struct _nfsStats {
    int nfsServer, nfsClient;
};
static struct _nfsStats nfsStats, old_nfsStats;

/*
 * the system_info struct is filled in by a machine dependent routine.
 */

struct system_info
{
    int    last_pid;
    double load_avg[NUM_AVERAGES];
    int    p_total;
    int    p_active;     /* number of procs considered "active" */
    int    *procstates;
    int    *cpustates;
    int    *memory;
};


/*
 * The array stats always has valid info for stats[i], 0 <= i < num_stats.
 * The info for the k-th possible statistic of interest is recorded, if it
 * is recorded at all, in stats[poss_stats[k]].
 */

int poss_stats[NUM_POSSIBLE_STATS];
#define WANT_STAT(x) (poss_stats[(x)] != NO_STAT)

stat_type stats;

static int num_stats;
static int kmem, mem;

#define max(a,b) (a>b ? a:b)

#if defined(PCS) && !defined(RCU) /* PCS M68 */
struct nlist nl[] = {
#define X_SYSINFO	0
	{ "_sysinfo" },
#define X_MAXFREE       1
	{ "_maxfree" },
#define X_FREEMEM	2
	{ "_freemem"},
#define X_BOOTTIME      3
	{ "_bootime" },
#define X_HZ		4
	{ "_hz" },
#define N_IFNET		5
	{ "_ifnet" },
	{ "" },
};
#else /* PCS && !RCU */
#if defined(PCS) && defined(RCU) /* PCS RCU */
struct nlist nl[] = {
#define X_SYSINFO	0
	{ "sysinfo" },
#define X_MAXFREE       1
	{ "maxfree" },
#define X_FREEMEM	2
	{ "freemem"},
#define X_BOOTTIME      3
	{ "bootime" },
#define X_HZ		4
	{ "hz" },
#define N_IFNET		5
	{ "ifnet" },
	{ "" },
};
#else /* PCS && RCU */
struct nlist nlst[] = {
#define X_CPTIME	0
	{ "_cpudata" },
#define X_RATE	  1
	{ "_rate" },
#define X_TOTAL	 2
	{ "_total" },
#define X_DEFICIT       3
	{ "_deficit" },
#define X_FORKSTAT      4
	{ "_forkstat" },
#define X_SUM	   5
	{ "_sum" },
#define X_FIRSTFREE     6
	{ "_firstfree" },
#define X_MAXFREE       7
	{ "_maxfree" },
#define X_BOOTTIME      8
	{ "_boottime" },
#define X_DKXFER	9
	{ "_dk_xfer" },
#define X_REC	   10
	{ "_rectime" },
#define X_PGIN	  11
	{ "_pgintime" },
#define X_HZ	    12
	{ "_hz" },
#define X_MBDINIT       13
	{ "_mbdinit" },
#define N_IFNET       14
	{ "_ifnet" },
#define X_UBDINIT	15
	{ "_ubdinit" },
#define X_IOCINIT	16
	{ "_ioccdinit" },
#define X_LOWCPU 17
	{ "_lowcpu" },
#define X_HIGHCPU 18
	{ "_highcpu" },
	{ "" },
};
#endif /* PCS && RCU */
#endif /* PCS && !RCU */

char dr_name[DK_NDRIVE][10];
char dr_unit[DK_NDRIVE];
double  stat1();
int     maxfree;
static long    hz;
static int     lowcpu = 0;
static int     highcpu = 0;

/* these are offsets obtained via nlist and used in the get_ functions */

static unsigned long avenrun_offset;
static unsigned long mpid_offset;
static unsigned long total_offset;
static unsigned long cpudata_offset;

/* these are for calculating cpu state percentages */

static long cp_time[CPUSTATES];
static long cp_old[CPUSTATES];
static long cp_diff[CPUSTATES];
static struct cpudata *cpudata[MAXCPU];
long percentages();

/* these are for detailing the cpu states */

int cpu_states[4];

#ifdef PCS
struct vmtotal
{
     long t_free;
};
#endif /* PCS */
struct {
	int     busy;
	long    time[CPUSTATES];
	long    xfer[DK_NDRIVE];
	struct  vmtotal Total;
#ifndef PCS
	struct  vmmeter Sum;
	struct  vmmeter Rate;
	struct  forkstat Forkstat;
	unsigned rectime;
	unsigned pgintime;
	long    interrupts[MAXCPU];
#endif /* !PCS */
} s, s1;
#define total s.Total
#define irpts s.interrupts
#ifndef PCS
#define rate s.Rate
#define sum  s.Sum
#define forksta	s.Forkstat

struct  vmmeter osum;
int     deficit;
#endif /* !PCS */
double  etime;
int     mf;
int     swflag;

int nintv;
long t;

int current_values[NUM_GRAPHS];

#define steal(where, var) lseek(kmem, where, 0); read(kmem, &var, sizeof var);
#define pgtok(a) ((a)*NBPG/1024)

int sys_setup()

/* This routine does all necessary setting up of structures
 * that will handle system calls.  It returns num_stats.
 */
{
    int stat;
    int have_disk;
    double atof();
  
    register int i = 0;
    register int pagesize;
    struct statics *statics;

    if ((kmem = open(KMEM, O_RDONLY)) == -1) {
        perror(KMEM);
        return(-1);
    }
    if ((mem = open(MEM, O_RDONLY)) == -1) {
        perror(MEM);
        return(-1);
    }

    /* get the list of symbols we want to access in the kernel */
    (void) nlist(VMUNIX, nlst);
    if (nlst[0].n_type == 0)
    {
        fprintf(stderr, "top: nlist failed\n");
        return(-1);
    }

    /* make sure they were all found */
    if (i > 0 && check_nlist(nlst) > 0)
    {
        return(-1);
    }

    (void) getkval(nlst[X_HZ].n_value,     (int *)(&hz),        sizeof(hz),
            nlst[X_HZ].n_name);
/*    (void) getkval(nlst[X_CCPU].n_value,   (int *)(&ccpu),      sizeof(ccpu),
            nlst[X_CCPU].n_name);*/
    (void) getkval(nlst[X_LOWCPU].n_value,   (int *)(&lowcpu),  sizeof(lowcpu),
            nlst[X_LOWCPU].n_name);
    (void) getkval(nlst[X_HIGHCPU].n_value,   (int *)(&highcpu),        sizeof(highcpu),
            nlst[X_HIGHCPU].n_name);

    /* stash away certain offsets for later use */

/*    avenrun_offset = nlst[X_AVENRUN].n_value;
    total_offset = nlst[X_TOTAL].n_value;*/
    cpudata_offset = nlst[X_CPTIME].n_value;

    nintv = get_namelist();

    collect_stats();
    etime = 1.0;
    have_disk = (total_disk_transfers() ? 1 : 0);
}

void update_stats()
{
    int *target[CPUSTATES-1], trash;

    collect_stats();
    current_values[USER_CPU_PERCENTAGE] = cpu_states[USER]/10.;
    current_values[SYSTEM_CPU_PERCENTAGE] = cpu_states[SYS]/10.;
    current_values[IDLE_CPU_PERCENTAGE] = cpu_states[IDLE]/10.;
    if (perfmon[FREE_MEM])
	current_values[FREE_MEM] = pgtok(total.t_free);
#ifndef PCS
    if (perfmon[DISK_TRANSFERS])
	current_values[DISK_TRANSFERS] = total_disk_transfers();
    if (perfmon[INPUT_PACKETS])
	current_values[INPUT_PACKETS] = packets.input - old_packets.input;
    if (perfmon[OUTPUT_PACKETS])
	current_values[OUTPUT_PACKETS] = packets.output - old_packets.output;
    if (perfmon[COLLISION_PACKETS])
	current_values[COLLISION_PACKETS] = packets.collisions - old_packets.collisions;
    if (perfmon[NFS_CLIENT_CALLS])
	current_values[NFS_CLIENT_CALLS] = nfsStats.nfsClient - old_nfsStats.nfsClient;
    if (perfmon[NFS_SERVER_CALLS])
	current_values[NFS_SERVER_CALLS] = nfsStats.nfsServer - old_nfsStats.nfsServer;
#endif /* !PCS */
}

#define	TIMER_EXPIRED(timer) \
    (*timer && ((*timer)->tv_sec == 0) && ((*timer)->tv_usec == 0))

int total_disk_transfers()
{
    register int i, total_xfers = 0;

    for(i=0; i < DK_NDRIVE; i++)
	total_xfers += s.xfer[i];
    return(total_xfers);
}

copy_cpu_stats(stat)
    int *stat[CPUSTATES-1];
{
    register int i;

    for(i=0; i<CPUSTATES; i++) {
	float f = stat1(i);
	if (i == 0) {           /* US+NI */
	    i++;
	    f += stat1(i);
	}
	if (stat[i-1] != 0)
	    *stat[i-1] = f;
    }
}

collect_stats()
{
    off_t ifnetaddr = (long)nlst[N_IFNET].n_value;
    register int i;
    long total_cpu;
    register int ncpu;
    register struct cpudata cpu;

    for (i = 0; i < CPUSTATES; ++i) {
        cp_time[i] = 0;
    }
    current_values[INTERRUPTS] = 0;
    (void) getkval(cpudata_offset, cpudata, sizeof(cpudata), "_cpudata");
    for (ncpu = lowcpu; ncpu <= highcpu; ++ncpu) {
        if (cpudata[ncpu] != NULL) {
	    (void) getkval(cpudata[ncpu], &cpu, sizeof(cpu), "???");
	    if (cpu.cpu_state & CPU_RUN) {
		for (i = 0; i < CPUSTATES; ++i) {
		    cp_time[i] += cpu.cpu_cptime[i];
		}
		if (perfmon[INTERRUPTS]) {                /* is the interrupts graph turned on? */
		    if ( !irpts[ncpu] )
			irpts[ncpu] = cpu.cpu_intr;	  /* for first time through */
		    else {
			current_values[INTERRUPTS] += cpu.cpu_intr - irpts[ncpu];
			irpts[ncpu] = cpu.cpu_intr;
		    }
		}
	    }
	}
    }

    /* convert cp_time counts to percentages */
    total_cpu = percentages(CPUSTATES, cpu_states, cp_time, cp_old, cp_diff);

    lseek(kmem, (long)nlst[X_DKXFER].n_value, 0);
    read(kmem, s.xfer, sizeof s.xfer);
    if (nintv != 1) {
	steal((long)nlst[X_SUM].n_value, rate);
    } else {
	steal((long)nlst[X_RATE].n_value, rate);
    }
    steal((long)nlst[X_TOTAL].n_value, total);
    osum = sum;
    steal((long)nlst[X_SUM].n_value, sum);
    steal((long)nlst[X_DEFICIT].n_value, deficit);

    etime = 1;
    for (i=0; i < DK_NDRIVE; i++) {
	t = s.xfer[i];
	s.xfer[i] -= s1.xfer[i];
	s1.xfer[i] = t;
    }

    if (nlst[N_IFNET].n_value != 0) {
	struct ifnet ifnet;
	steal((long)nlst[N_IFNET].n_value, ifnetaddr);
	old_packets = packets;
	packets.input = packets.output = packets.collisions = 0;
	while (ifnetaddr) {
	    steal(ifnetaddr, ifnet);
	    packets.input += ifnet.if_ipackets;
	    packets.output += ifnet.if_opackets;
	    packets.collisions += ifnet.if_collisions;
	    ifnetaddr = (off_t) ifnet.if_next;
	}
    }
    getnfs(FALSE,TRUE,TRUE,TRUE);
    old_nfsStats = nfsStats;
    nfsStats.nfsServer = rpcNnfs.nfs_s.ncalls;
    nfsStats.nfsClient = rpcNnfs.nfs_c.ncalls;
}

min (a, b)
    int a,b;
{
    return(a<b ? a:b);
}
int check_nlist(nlst)

register struct nlist *nlst;

{
    register int i;

    /* check to see if we got ALL the symbols we requested */
    /* this will write one line to stderr for every symbol not found */

    i = 0;
    while (nlst->n_name != NULL)
    {
        if (nlst->n_type == 0)
        {
            /* this one wasn't found */
            fprintf(stderr, "kernel: no symbol named `%s'\n", nlst->n_name);
            i = 1;
        }
        nlst++;
    }

    return(i);
}

int get_namelist()

{
    time_t now;
    time_t boottime;
    register int i;
    int nintv;

    steal((long)nlst[X_MAXFREE].n_value, maxfree);
    steal((long)nlst[X_BOOTTIME].n_value, boottime);
    steal((long)nlst[X_HZ].n_value, hz);
    for (i = 0; i < DK_NDRIVE; i++) {
	strcpy(dr_name[i], "xx");
	dr_unit[i] = i;
    }
#ifndef PCS
    /* I don't see a way to obtain the information that
     * read_names gets so it is commented it out for now.
     */
    read_names();
#endif /* !PCS */
    time(&now);
    nintv = now - boottime;
    if (nintv <= 0 || nintv > 60*60*24*365*10) {
	fprintf(stderr,
	    "Time makes no sense... namelist must be wrong.\n");
	exit(1);
    }
    return(nintv);
}

double
stat1(row)
{
    double t;
    register i;

    t = 0;
    for(i=0; i<CPUSTATES; i++)
	t += s.time[i];
    if(t == 0.)
	t = 1.;
    return(s.time[row]*100./t);
}

/*
 *  percentages(cnt, out, new, old, diffs) - calculate percentage change
 *      between array "old" and "new", putting the percentages i "out".
 *      "cnt" is size of each array and "diffs" is used for scratch space.
 *      The array "old" is updated on each call.
 *      The routine assumes modulo arithmetic.  This function is especially
 *      useful on BSD mchines for calculating cpu state percentages.
 */

long percentages(cnt, out, new, old, diffs)

int cnt;
int *out;
register long *new;
register long *old;
long *diffs;

{
    register int i;
    register long change;
    register long total_change;
    register long *dp;
    long half_total;

    /* initialization */
    total_change = 0;
    dp = diffs;

    /* calculate changes for each state and the overall change */
    for (i = 0; i < cnt; i++)
    {
        if ((change = *new - *old) < 0)
        {
            /* this only happens when the counter wraps */
            change = (int)
                ((unsigned long)*new-(unsigned long)*old);
        }
        total_change += (*dp++ = change);
        *old++ = *new++;
    }

    /* calculate percentages based on overall change, rounding up */
    half_total = total_change / 2l;
    for (i = 0; i < cnt; i++)
    {
        *out++ = (int)((*diffs++ * 1000 + half_total) / total_change);
    }

    /* return the total in case the caller wants to use it */
    return(total_change);
}

#ifdef vax
read_names()
{
    struct mba_device mdev;
    register struct mba_device *mp;
    struct mba_driver mdrv;
    short two_char;
    char *cp = (char *) &two_char;
    struct uba_device udev, *up;
    struct uba_driver udrv;

    mp = (struct mba_device *) nlst[X_MBDINIT].n_value;
    up = (struct uba_device *) nlst[X_UBDINIT].n_value;
    if (up == 0) {
	fprintf(stderr, "perfmon: Disk init info not in namelist\n");
	exit(1);
    }
    if(mp) for (;;) {
	steal(mp++, mdev);
	if (mdev.mi_driver == 0)
	    break;
	if (mdev.mi_dk < 0 || mdev.mi_alive == 0)
	    continue;
	steal(mdev.mi_driver, mdrv);
	steal(mdrv.md_dname, two_char);
	sprintf(dr_name[mdev.mi_dk], "%c%c", cp[0], cp[1]);
	dr_unit[mdev.mi_dk] = mdev.mi_unit;
    }
    if(up) for (;;) {
	steal(up++, udev);
	if (udev.ui_driver == 0)
	    break;
	if (udev.ui_dk < 0 || udev.ui_alive == 0)
	    continue;
	steal(udev.ui_driver, udrv);
	steal(udrv.ud_dname, two_char);
	sprintf(dr_name[udev.ui_dk], "%c%c", cp[0], cp[1]);
	dr_unit[udev.ui_dk] = udev.ui_unit;
    }
}
#endif vax

#ifdef sun
read_names()
{
   struct mb_device mdev;
   register struct mb_device *mp;
   struct mb_driver mdrv;
   short two_char;
   char *cp = (char *) &two_char;

   mp = (struct mb_device *) nlst[X_MBDINIT].n_value;
   if (mp == 0) {
	fprintf(stderr, "vmstat: Disk init info not in namelist\n");
	return(0);
   }
   for (;;) {
	steal(mp++, mdev);
	if (mdev.md_driver == 0)
	    break;
	if (mdev.md_dk < 0 || mdev.md_alive == 0)
	    continue;
	steal(mdev.md_driver, mdrv);
	steal(mdrv.mdr_dname, two_char);
	sprintf(dr_name[mdev.md_dk], "%c%c", cp[0], cp[1]);
	dr_unit[mdev.md_dk] = mdev.md_unit;
   }
}
#endif sun

#ifdef ultrix
read_names()
{
    return;
}
#endif

#ifdef ibm032
read_names()
{
    struct iocc_device mdev;
    register struct iocc_device *mp;
    struct iocc_driver mdrv;
    short two_char;
    char *cp = (char *) &two_char;

    mp = (struct iocc_device *) nlst[X_IOCINIT].n_value;
    if (mp == 0) {
	fprintf(stderr, "vmstat: Disk init info not in namelist\n");
	return(0);
    }
    for (;;) {
	steal(mp++, mdev);
	if (mdev.iod_driver == 0)
	    break;
	if (mdev.iod_dk < 0 || mdev.iod_alive == 0)
	    continue;
	steal(mdev.iod_driver, mdrv);
	steal(mdrv.idr_dname, two_char);
	sprintf(dr_name[mdev.iod_dk], "%c%c", cp[0], cp[1]);
	dr_unit[mdev.iod_dk] = mdev.iod_unit;
    }
}
#endif ibm032

/*
 *  getkval(offset, ptr, size, refstr) - get a value out of the kernel.
 *      "offset" is the byte offset into the kernel for the desired value,
 *      "ptr" points to a buffer into which the value is retrieved,
 *      "size" is the size of the buffer (and the object to retrieve),
 *      "refstr" is a reference string used when printing error meessages,
 *          if "refstr" starts with a '!', then a failure on read will not
 *          be fatal (this may seem like a silly way to do things, but I
 *          really didn't want the overhead of another argument).
 *  
 */

getkval(offset, ptr, size, refstr)

unsigned long offset;
int *ptr;
int size;
char *refstr;

{
    if (lseek(kmem, (long)offset, L_SET) == -1) {
        if (*refstr == '!')
            refstr++;
        (void) fprintf(stderr, "%s: lseek to %s: %s\n", KMEM, 
                       refstr, strerror(errno));
        exit(23);
    }
    if (read(kmem, (char *) ptr, size) == -1) {
        if (*refstr == '!') 
            return(0);
        else {
            (void) fprintf(stderr, "%s: reading %s: %s\n", KMEM, 
                           refstr, strerror(errno));
            exit(23);
        }
    }
    return(1);
}
